/*
  yasep/defora/gen_asm.js created by whygee at f-cpu.org
  This file is distributed under terms of the Affero GPL v3 license or later, see http://yasep.org
  created 2011-09-25 from yasep/tool/test_opcodes.js
  version lun. nov.  7 13:21:06 CET 2011 : changed name, changed for use with filefox
  version 2011-11-28 : big refactoring, creates registers
  version 2012-08-13 : ASM.YASEP_TYPE=32;  => IMM20 works

  This menu item creates a dump of the opcodes in all forms for defora's assembler.
    2 outputs are provided :
       - an assembly testbench/regression file, with source/cleartext and assembled/hexa versions
       - another dump for creating the assembler's parser rules.
*/

init_obj.push(DEF_ASM={
  win:null,

  make_reg: function(){
    var h="/* defora/src/arch/yasep.reg\n"
         +"   Copyright (c) 2011-2012 Yann Guidon <whygee@f-cpu.org>\n"
         +"    derived from code created by Pierre Pronchery <khorben@defora.org>\n"
         +"   This file is part of DeforaOS Devel asm\n"
         +    generated("  ","defora/gen_asm.js")
         +" */\n\n";

    for (var i in Y.registers_names) {
      h+="REG("+pad_right(Y.registers_names[i].toLowerCase(),4)
          +", W, "+pad_left(i,3)+")\n";
    }
    savefile.save("yasep.reg", h);
  },


  make_S: function(){
    var errnum=0,
        msg_number=0,
        h="/* DeforaOS/Apps/Devel/src/asm/test/yasep.S\n"
         +"   Copyright (c) 2011 Pierre Pronchery <khorben@defora.org>\n"
         +"   Copyright (c) 2011-2012 Yann Guidon <whygee@f-cpu.org>\n"
         +"   This file is part of DeforaOS Devel asm\n\n"
         +"   Assembly language dump to test the Defora\n"
         +"    assembler for the YASEP architecture.\n"
         +"   Bytes are in little endian order.\n\n"
         +    generated("  ","defora/gen_asm.js")
         +" */\n\n.text\n";

    function add_message(col,msg){
      if ((col=="blue") || (col=="green"))
        return;
      h +=" ###### "+msg+"*/\n";
    }

    var ASM = new Y.ASM_context(add_message);
    ASM.YASEP_TYPE=32; // => IMM20 works

    function test_opcode(x) {
      h+="\t"+pad_right(x,24).toLowerCase()+"/* ";
      if (Y.encode_instruction(x,ASM)) {
        var e=ASM.encoded_instruction,
            s=ASM.encoded_size_bytes,
            t=" ";
        while (s--) { // displays every byte, little endian style
          t+=pad_left("0x"+int2hex(e&255,1,1,1),5);
          e>>=8;
        }
        h+=pad_right(ASM.form,10)+pad_right(t,25)+" */\n";
      }
      else
        errnum++;
    }

    for (var cur_opcode in DEF_ASM.sorted_opcodes){
      var name=DEF_ASM.sorted_opcodes[cur_opcode],
          instr=Y.table_opcode[name];
          index=instr.val,
          f=instr.forms,
          g=instr.flags,
          i=1,
          j="",
          a=Y.opcode_aliases[name];

      if (g & Y.YASEP16_ONLY) {
        h+="#if defined(ARCH_yasep16)\n";
        j="#endif\n";
      }
      else
        if (g & Y.YASEP32_ONLY) {
          h+="#if defined(ARCH_yasep32)\n";
          j="#endif\n";
        }

      if (a) {
        h+="#if defined(ALLOW_YASEP_ALIASES)\n";
        j+="#endif\n";
      }

      h+="/* 0x"+(int2hex(index,1,1,1))+" = "+name;
      if (a)
        h+=" (alias to "+a+")";
      h+=" : "+instr.description+"*/\n";

      // Scan the forms list
      do {
        if (f & i) {
          var w=Y.TableEncoders[Y.TableForm[i]].ex(g,name); // random operands
          if (w=="") // no operands so don't add a spacer that will bork the instruction
            w=name;
          else 
            w=name+" "+w;

          test_opcode(w);

          if ((i & Y.FORMS_EXTENDABLE)&&((g & Y.NO_CONDITION)==0))
            test_opcode(w+" "+Y.randomCondition());
        }
        i<<=1;
      } while (i <= f);

      h+=j+"\n";
    } // for

    savefile.save("yasep.S", h);
  },

  make_flags: function(){
    var i, j, t, h=
"/* DeforaOS/Apps/Devel/src/asm/src/arch/yasep.flags\n"+
"   Copyright (c) 2011 Yann Guidon <whygee@f-cpu.org>\n"+
"   This file is part of DeforaOS Devel asm\n"+
    generated("  ","defora/gen_asm.js")+
" */\n\n";

    // list all the flags
    for (i in Y.TableFlags) {
      h+="#define "+pad_right(Y.TableFlags[i],15)+pad_left(i,8)+"\n";
    }

    h+="\n\nuint32_t _yasep_flags[64]={\n";
    i=0;
    while (true) {
      h+="0";
      if (t=Y.opcode_table[i<<2]) {
        h+=" /* "+pad_right(t.name,6)+"*/";
        j=1;
        t=t.flags;

        while (j<=t) {
          if (j & t)
            h+=" | "+Y.TableFlags[j];
          j<<=1;
        }
      }

      i++;
      if (i>=64)
        break;
      h+=",\n";
    }
    h+="\n};\n\n";

    // list all the formss
    for (i in Y.TableForm) {
      h+="#define "+pad_right(Y.TableForm[i],11)+pad_left(i,6)+"\n";
    }

    h+="\n\nuint32_t _yasep_forms[64]={\n";
    i=0;
    while (true) {
      h+="0";
      if (t=Y.opcode_table[i<<2]) {
        h+=" /* "+pad_right(t.name,6)+"*/";
        j=1;
        t=t.forms;

        while (j<=t) {
          if (j & t)
            h+=" | "+Y.TableForm[j];
          j<<=1;
        }
      }

      i++;
      if (i>=64)
        break;
      h+=",\n";
    }
    h+="\n}\n";

    savefile.save("yasep.flags", h);
  },

  make_ins: function(){
    var i, j=0, k=0, h=
"/* DeforaOS/Apps/Devel/src/asm/src/arch/yasep.ins\n"+
"   Copyright (c) 2011 Yann Guidon <whygee@f-cpu.org>\n"+
"   Derived from :\n"+
"   Copyright (c) 2011 Pierre Pronchery <khorben@defora.org>\n"+
"   This file is part of DeforaOS Devel asm\n"+
    generated("  ","defora/gen_asm.js")+" */\n\n"+
"#define OP_REG    AO_REGISTER(0,  W,  0)\n"+
"#define OP_IMM4   AO_IMMEDIATE(0, 4,  0)\n"+
"#define OP_IMM5   AO_IMMEDIATE(0, 5,  0)\n"+
"#define OP_IMM8   AO_IMMEDIATE(0, 8,  0)\n"+
"#define OP_IMM16  AO_IMMEDIATE(0, 16, 0)\n"+
"#define OP_IMM20  AO_IMMEDIATE(0, 20, 0)\n"+
"\n";

    // list all the forms
    for (i in Y.TableForm)
      h+="#define "+pad_right(Y.TableForm[i]+"_F",12)+pad_left(j++,3)+"|(16 << AOD_SIZE)\n";
    h+="\n";

    for (var cur_opcode in DEF_ASM.sorted_opcodes){
      var name=DEF_ASM.sorted_opcodes[cur_opcode],
          instr=Y.table_opcode[name],
          index=instr.val,
          s,
          t,
          f=instr.forms,
          g=instr.flags,
          i=1,
          j="",
          a=Y.opcode_aliases[name];

      if (g & Y.YASEP16_ONLY) {
        h += "#if defined(ARCH_yasep16)\n";
        j="#endif\n";
      }
      else
        if (g & Y.YASEP32_ONLY) {
          h += "#if defined(ARCH_yasep32)\n";
          j="#endif\n";
        }

      if (a) {
        h+="#if defined(ALLOW_YASEP_ALIASES)\n";
        j+="#endif\n";
      }

      h+='/**** '+name;
      if (a)
        h+=" (alias to "+a+")";
      h+=' ****/\n';

      // Scan the forms list
      do {
        t=Y.TableForm[i];
        if ((f & i) && Y.TableEncoders[t].de) {
          s=Y.TableEncoders[t].lsb(g);
          h+="{ "+pad_right('"'+name.toLowerCase()+'", ',9)
           + "0x"+int2hex(index|s,1,1,1)
           + ", "+Y.TableEncoders[t].de(g)+" },\n";
/*
          if ((i & Y.FORMS_EXTENDABLE)&&((g & Y.NO_CONDITION)==0)) {
            z="* "+z;
            y=" } not implemented yet /\n";
            h+=z+",  Y_COND_ALONE, null  "+y
             + z+",  Y_COND_REG,   OP_R  "+y
             + z+",  Y_COND_IM4,   IMM4  "+y;
          }
*/
        }
        i<<=1;
      } while (i <= f);
      h+=j+"\n";
    }

    savefile.save("yasep.ins", h.slice(0,-2)); // Trim the last ",\n"
  },


///// triggers


  start: function() {
    // prepare the sorted list of opcodes and aliases
    DEF_ASM.sorted_opcodes=[];
    for (var name in Y.table_opcode) {
      if (Y.table_opcode[name].forms>0)
        DEF_ASM.sorted_opcodes.push(name);
    }
    DEF_ASM.sorted_opcodes=DEF_ASM.sorted_opcodes.sort();

    var div = dcE("DIV"),
        c=I8N("c")+" ",
        b='</button><br><button id="';
    div.innerHTML='<button id="save_S">'+c+'yasep.S'+b
         +'save_ins">'+c+'yasep.ins'+b
         +'save_flags">'+c+'yasep.flags'+b
         +'save_reg">'+c+'yasep.reg</button>';

    getById("save_S",null,div).onclick=DEF_ASM.make_S;
    getById("save_ins",null,div).onclick=DEF_ASM.make_ins;
    getById("save_reg",null,div).onclick=DEF_ASM.make_reg;
    getById("save_flags",null,div).onclick=DEF_ASM.make_flags;

    return(DEF_ASM.win= ygwm.new_window("Defora ASM files", null, div,
       150, 120, 600, 70, false, true, true, true, false, "doc_gryl"));
  },

  init_obj: function() {
    menu.subMenuItem(menu.list["Dev"], i8n.def, "def/asm");
  }
});
